local _, Data = ...
local NameplateCCnTrinket = LibStub("AceAddon-3.0"):NewAddon("NameplateCCnTrinket", "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("NameplateCCnTrinket")
local LCG = LibStub("LibCustomGlow-1.0")
local DRL = LibStub("DRList-1.0")

local _G, pairs, select, band, floor, strfind = _G, pairs, select, bit.band, math.floor, string.find
local CreateFrame, CreateTexture = CreateFrame, CreateTexture
local GetSpellInfo, GetTime = GetSpellInfo, GetTime
local UnitAura, UnitGUID = UnitAura, UnitGUID
local CooldownFrame_Set = CooldownFrame_Set
local CombatLogGetCurrentEventInfo = CombatLogGetCurrentEventInfo
local C_NamePlate, C_CVar = C_NamePlate, C_CVar
local GetCVar, GetNamePlateForUnit = C_CVar.GetCVar, C_NamePlate.GetNamePlateForUnit
local COMBATLOG_OBJECT_REACTION_HOSTILE, COMBATLOG_OBJECT_REACTION_MASK = COMBATLOG_OBJECT_REACTION_HOSTILE, COMBATLOG_OBJECT_REACTION_MASK

local DR_TIME = DRL.resetTimes.retail["default"]
local getUpdate = 0
local CC_MODE = 0
local CD_MODE = 1
local npVisible = {}
local npUnitID = {}
local CDTimeCache = {}
local CDTextureCache = {}
local gradual = {}
local cooldown = {}
local UPDATE_INTERVAL = 0.3
local PGUID = UnitGUID("player")
local CommonIcon = {
	["taunt"] 			= 355,
	["incapacitate"]	= 118,
	["silence"] 		= 15487,
	["disorient"] 		= 118699,
	["stun"] 			= 408,
	["root"] 			= 122,
	["knockback"] 		= 132469,
	["disarm"] 			= 236077,
}
local DefaultConfig = {
	profile = {
		gSetting = {
			ShowFriendlyPlayer = true,
			CCCommonIcon = false,
			CCShowMonster = false,
			CurrentTime = true,
			CooldownSpiral = true,
			FrameSize = 25,
			LeftxOfs = 0,
			RightxOfs = 0,
			yOfs = 0,
			TargetAlpha = 1,
			OtherAlpha = 0.6,
			OtherScale = 1 / GetCVar("nameplateSelectedScale"),
		},
		pSetting = {
			pEnable = true,
			pxOfs = 0,
			pyOfs = 0,
			pScale = 0.9,
			attachFrame = "PlayerFrame",
		},
		Func = {
			Interrupt = true,
			Racial = true,
			Trinket = true,
			CC = true,
			Dispel = true,
		},
		Group = {
			taunt = false,
			incapacitate = true,
			silence = true,
			disorient = true,
			stun = true,
			root = true,
			knockback = false,
			disarm = true,
		},
	}
}

function NameplateCCnTrinket:OnInitialize()
	-- Called when the addon is loaded
	self.Settings = LibStub("AceDB-3.0"):New("NameplateCCnTrinketSettings", DefaultConfig, true)

	self.Settings.RegisterCallback(self, "OnProfileChanged", "Refresh")
	self.Settings.RegisterCallback(self, "OnProfileCopied", "Refresh")
	self.Settings.RegisterCallback(self, "OnProfileReset", "Refresh")
	self.Settings.RegisterCallback(self, "OnProfileShutdown", "Refresh")

	self:Option()
	self:RegisterChatCommand("nct", "ChatCommand")
	self:RegisterChatCommand("NameplateCCnTrinket", "ChatCommand")

	for str in pairs(Data.n_MAP) do
		for sp, tm in pairs(Data.n_MAP[str]) do
			if select(1, GetSpellInfo(sp)) == nil then
				DEFAULT_CHAT_FRAME:AddMessage("|c00008000NameplateCCnTrinket|r n_Map["..str.."] ".. sp)
			else
				CDTimeCache[sp] = tm
				CDTextureCache[sp] = select(3, GetSpellInfo(sp))
			end
		end
	end
	CDTextureCache[195901] = "Interface\\Icons\\Sha_ability_rogue_sturdyrecuperate" -- Adaptation
	--CDTextureCache[196029] = "Interface\\Icons\\Ability_bossdarkvindicator_auraofcontempt" -- Relentless
	
	for str, id in pairs(CommonIcon) do
		if select(1, GetSpellInfo(id)) == nil then
			DEFAULT_CHAT_FRAME:AddMessage("|c00008000NameplateCCnTrinket|r [CommonIcon] " .. id)
		end
	end

	self.Frame = CreateFrame("Frame")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("NAME_PLATE_UNIT_ADDED")
	self:RegisterEvent("NAME_PLATE_UNIT_REMOVED")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self.Frame:SetScript("OnUpdate", self.OnUpdate)
end

function NameplateCCnTrinket:Refresh()
	self:ClearValue()
end

function NameplateCCnTrinket:ChatCommand(input)
	if not input or input:trim() == "" then
		LibStub("AceConfigDialog-3.0"):SetDefaultSize("NameplateCCnTrinket", 600, 470)
		LibStub("AceConfigDialog-3.0"):Open("NameplateCCnTrinket")
	else
		LibStub("AceConfigCmd-3.0"):HandleCommand("nct", "NameplateCCnTrinket", input)
	end
end

local function GetAuraDuration(unitID, spellID)
	if not unitID then return end

	for i = 1, 40 do
		local _, _, _, _, duration, _, _, _, _, id = UnitAura(unitID, i, "HARMFUL")
		if not id then return end -- no more debuffs

		if spellID == id then
			return duration
		end
	end
end

local function isHarm(fl)
	if NameplateCCnTrinket.Settings.profile.gSetting.ShowFriendlyPlayer then
		return band(fl, COMBATLOG_OBJECT_REACTION_MASK)
	else
		return band(fl, COMBATLOG_OBJECT_REACTION_HOSTILE)
	end
end

local function isTarget(n, tar)
	local tn = UnitGUID(tar)
	return (tn and n == tn)
end

local function CreateBorderTexture(FirstName, SecondName)
	local tframe = CreateFrame("Frame", FirstName .. SecondName)
	tframe:SetFrameStrata("BACKGROUND")
	tframe:SetSize(NameplateCCnTrinket.Settings.profile.gSetting.FrameSize, NameplateCCnTrinket.Settings.profile.gSetting.FrameSize)
	tframe:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)

	tframe.Texture = tframe:CreateTexture(nil, "BACKGROUND")
	tframe.Texture:SetAllPoints()
	local ctex = tframe:CreateTexture(nil, "OVERLAY")
	ctex:SetTexture(1, 1, 1)

	tframe.border = tframe:CreateTexture(nil, "BORDER")
	tframe.border:SetTexture("Interface\\Buttons\\UI-Debuff-Overlays")
	tframe.border:SetTexCoord(0.296875, 0.5703125, 0, 0.515625)
	tframe.border:SetVertexColor(0, 0, 0)
	tframe.border:SetAllPoints()

	tframe.c = CreateFrame("Cooldown", nil, tframe, "CooldownFrameTemplate")
	tframe.c:SetFrameLevel(0)
	tframe.c:SetSwipeTexture(ctex:GetTexture())
	tframe.c:SetSwipeColor(0, 0, 0, 0.6)
	tframe.c:SetReverse(true)
	tframe.c:SetDrawSwipe(NameplateCCnTrinket.Settings.profile.gSetting.CooldownSpiral)
	tframe.c:SetHideCountdownNumbers(false)
	tframe.c:SetAllPoints()
end

local function CreateDiminishFrame(tempGUID, tempSpellID, isApplied, isTest)
	local id, cat = NameplateCCnTrinket:CheckCategory(tempSpellID)
	if id == nil then return end

	if not gradual[tempGUID] then gradual[tempGUID] = {} end
	if not gradual[tempGUID][cat] then
		CreateBorderTexture(cat, tempGUID)
		gradual[tempGUID][cat] = _G[cat .. tempGUID]
		_G[cat .. tempGUID].c:SetScript("OnHide", function(self) self.count = nil end)
	end

	local tempFrame = _G[cat .. tempGUID]
	local fTime

	if isApplied then
		if tempFrame.c.count == nil then
			tempFrame.c.count = 1
		else
			tempFrame.c.count = tempFrame.c.count + 1
		end

		if isTest then
			fTime = 0
		else
			local tmpuid
			if tempGUID == PGUID then
				tmpuid = "player"
			else
				tmpuid = npUnitID[tempGUID]
			end
			fTime = GetAuraDuration(tmpuid, tempSpellID)

			if fTime == nil then
				return
			end
		end
	else
		if tempFrame.c.count == nil then
			tempFrame.c.count = 0
		end
		fTime = 0
	end

	local mask_rgb = { 0, 1, 0, 0.6 }
	if tempFrame.c.count == 2 then
		mask_rgb = { 1, 1, 0, 0.6 }
	elseif tempFrame.c.count > 2 then
		mask_rgb = { 1, 0, 0, 0.6 }
	end

	if NameplateCCnTrinket.Settings.profile.gSetting.CurrentTime then
		if isApplied then
			LCG.ButtonGlow_Start(tempFrame.c, mask_rgb)
		else
			LCG.ButtonGlow_Stop(tempFrame.c)
		end
	end

	tempFrame.border:SetVertexColor(mask_rgb[1], mask_rgb[2], mask_rgb[3])
	if NameplateCCnTrinket.Settings.profile.gSetting.CCCommonIcon then
		tempFrame.Texture:SetTexture(select(3, GetSpellInfo(id)))
	else
		tempFrame.Texture:SetTexture(select(3, GetSpellInfo(tempSpellID)))
	end
	tempFrame.c:SetCooldown(GetTime(), fTime + DR_TIME)
end

local function HideFrame(frame)
	if frame then
		frame:Hide()
		frame.c:Hide()
		frame.border:Hide()
	end
end

local function UpdateFrame(g_tb, sel)
	for n, tb in pairs(g_tb) do
		if NameplateCCnTrinket.Settings.profile.pSetting.pEnable and sel == CC_MODE and n == PGUID then
			local alpha, scale, gn = 0, 0, 0

			alpha = NameplateCCnTrinket.Settings.profile.gSetting.TargetAlpha
			scale = NameplateCCnTrinket.Settings.profile.pSetting.pScale

			for id, fr in pairs(tb) do
				fr:ClearAllPoints()
				if fr.c:IsVisible() then
					fr:SetAlpha(alpha)
					fr:SetScale(scale)
					fr:SetPoint("TOPLEFT", NameplateCCnTrinket.Settings.profile.pSetting.attachFrame, "TOPLEFT", NameplateCCnTrinket.Settings.profile.pSetting.pxOfs + (NameplateCCnTrinket.Settings.profile.gSetting.FrameSize + 2) * gn, NameplateCCnTrinket.Settings.profile.pSetting.pyOfs + NameplateCCnTrinket.Settings.profile.gSetting.FrameSize * 2)
					gn = gn + 1
				else
					fr:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)
				end
			end
		else
			local pl
			for frame, unitName in pairs(npVisible) do
				if n == unitName then
					pl = frame
					break
				end
			end

			if pl then
				local alpha, scale, gn = 0, 0, 0

				if isTarget(n, "target") or isTarget(n, "focus") then
					alpha = NameplateCCnTrinket.Settings.profile.gSetting.TargetAlpha
					scale = 1
				else
					alpha = NameplateCCnTrinket.Settings.profile.gSetting.OtherAlpha
					scale = NameplateCCnTrinket.Settings.profile.gSetting.OtherScale
				end

				for id, fr in pairs(tb) do
					fr:ClearAllPoints()
					if fr.c:IsVisible() then
						fr:SetAlpha(alpha)
						fr:SetScale(scale)
						if sel == CD_MODE then
							if Data.n_MAP["Trinket"][id] then
								if NameplateCCnTrinket.Settings.profile.Func.Trinket then
									fr:SetPoint("BOTTOMLEFT", pl, "TOPRIGHT", NameplateCCnTrinket.Settings.profile.gSetting.RightxOfs, NameplateCCnTrinket.Settings.profile.gSetting.yOfs - NameplateCCnTrinket.Settings.profile.gSetting.FrameSize)
								else
									fr:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)
								end
							elseif Data.n_MAP["Racial"][id] then
								if NameplateCCnTrinket.Settings.profile.Func.Racial then
									fr:SetPoint("BOTTOMLEFT", pl, "TOPRIGHT", NameplateCCnTrinket.Settings.profile.gSetting.RightxOfs, NameplateCCnTrinket.Settings.profile.gSetting.yOfs - NameplateCCnTrinket.Settings.profile.gSetting.FrameSize * 2)
								else
									fr:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)
								end
							elseif Data.n_MAP["Interrupt"][id] then
								if NameplateCCnTrinket.Settings.profile.Func.Interrupt then
									fr:SetPoint("BOTTOMRIGHT", pl, "TOPLEFT", NameplateCCnTrinket.Settings.profile.gSetting.LeftxOfs, NameplateCCnTrinket.Settings.profile.gSetting.yOfs - NameplateCCnTrinket.Settings.profile.gSetting.FrameSize)
								else
									fr:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)
								end
							elseif Data.n_MAP["Dispel"][id] then
								if NameplateCCnTrinket.Settings.profile.Func.Dispel then
									fr:SetPoint("BOTTOMRIGHT", pl, "TOPLEFT", NameplateCCnTrinket.Settings.profile.gSetting.LeftxOfs, NameplateCCnTrinket.Settings.profile.gSetting.yOfs - NameplateCCnTrinket.Settings.profile.gSetting.FrameSize * 2)
								else
									fr:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)
								end
							end
						elseif sel == CC_MODE then
							if NameplateCCnTrinket.Settings.profile.Func.CC then
								fr:SetPoint("BOTTOMLEFT", pl, "TOPRIGHT", 2 + NameplateCCnTrinket.Settings.profile.gSetting.RightxOfs + NameplateCCnTrinket.Settings.profile.gSetting.FrameSize + NameplateCCnTrinket.Settings.profile.gSetting.FrameSize * floor(gn / 2), NameplateCCnTrinket.Settings.profile.gSetting.yOfs - NameplateCCnTrinket.Settings.profile.gSetting.FrameSize - NameplateCCnTrinket.Settings.profile.gSetting.FrameSize * (gn % 2))
								gn = gn + 1
							else
								fr:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)
							end
						end
					else
						fr:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)
					end
				end
			else
				for _, fr in pairs(tb) do
					fr:ClearAllPoints()
					fr:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", -100, 0)
				end
			end
		end
	end
end

function NameplateCCnTrinket:ClearValue()
	for n, table in pairs(cooldown) do
		for id, frame in pairs(table) do
			HideFrame(frame)
			_G[n .. id].c = nil
			_G[n .. id] = nil
		end
	end
	for n, table in pairs(gradual) do
		for id, frame in pairs(table) do
			HideFrame(frame)
			_G[id .. n].c = nil
			_G[id .. n] = nil
		end
	end
	gradual = {}
	cooldown = {}
	npUnitID = {}
	--npVisible = {}
end

function NameplateCCnTrinket:Test()
	self:ClearValue()

	local GUID = UnitGUID("target")
	if GUID == nil then
		DEFAULT_CHAT_FRAME:AddMessage(L["selectnameplate"])
		return
	end
	local PUID = PGUID
	local spellID = { 208683, 59752, 6552, 527 }
	local testset = true
	local ct = self.Settings.profile.gSetting.CurrentTime
	if not cooldown[GUID] then cooldown[GUID] = {} end
	for i = 1, #spellID do
		CreateBorderTexture(GUID, spellID[i])
		_G[GUID .. spellID[i]].Texture:SetTexture(CDTextureCache[spellID[i]])
		cooldown[GUID][spellID[i]] = _G[GUID .. spellID[i]]
		cooldown[GUID][spellID[i]].c:SetCooldown(GetTime(), CDTimeCache[spellID[i]])
	end

	if self.Settings.profile.Group.taunt then
		CreateDiminishFrame(GUID, CommonIcon["taunt"], ct, testset)
		CreateDiminishFrame(PUID, CommonIcon["taunt"], ct, testset)
	end
	if self.Settings.profile.Group.incapacitate then
		CreateDiminishFrame(GUID, CommonIcon["incapacitate"], ct, testset)
		CreateDiminishFrame(PUID, CommonIcon["incapacitate"], ct, testset)
	end
	if self.Settings.profile.Group.silence then
		CreateDiminishFrame(GUID, CommonIcon["silence"], ct, testset)
		CreateDiminishFrame(PUID, CommonIcon["silence"], ct, testset)
	end
	if self.Settings.profile.Group.disorient then
		CreateDiminishFrame(GUID, CommonIcon["disorient"], ct, testset)
		CreateDiminishFrame(PUID, CommonIcon["disorient"], ct, testset)
	end
	if self.Settings.profile.Group.stun then
		CreateDiminishFrame(GUID, CommonIcon["stun"], ct, testset)
		CreateDiminishFrame(PUID, CommonIcon["stun"], ct, testset)
	end
	if self.Settings.profile.Group.root then
		CreateDiminishFrame(GUID, CommonIcon["root"], ct, testset)
		CreateDiminishFrame(PUID, CommonIcon["root"], ct, testset)
	end
	if self.Settings.profile.Group.knockback then
		CreateDiminishFrame(GUID, CommonIcon["knockback"], ct, testset)
		CreateDiminishFrame(PUID, CommonIcon["knockback"], ct, testset)
	end
	if self.Settings.profile.Group.disarm then
		CreateDiminishFrame(GUID, CommonIcon["disarm"], ct, testset)
		CreateDiminishFrame(PUID, CommonIcon["disarm"], ct, testset)
	end
end

function NameplateCCnTrinket:CheckCategory(spellID)
	--DRData.spells[spell] = Category ex) DRData.spells[339] root
	--DRData.providers[spell] = provider ex) DRData.providers[339] 339
	local  tempstr = DRL.spells[spellID]
	if     tempstr == "taunt"       and self.Settings.profile.Group.taunt        then
	elseif tempstr == "incapacitate"and self.Settings.profile.Group.incapacitate then
	elseif tempstr == "silence"     and self.Settings.profile.Group.silence      then
	elseif tempstr == "disorient"   and self.Settings.profile.Group.disorient    then
	elseif tempstr == "stun"        and self.Settings.profile.Group.stun         then
	elseif tempstr == "root"        and self.Settings.profile.Group.root         then
	elseif tempstr == "knockback"   and self.Settings.profile.Group.knockback    then
	elseif tempstr == "disarm"      and self.Settings.profile.Group.disarm       then
	else return nil, nil
	end
	return CommonIcon[tempstr], tempstr
end

function NameplateCCnTrinket:PLAYER_ENTERING_WORLD()
	self:ClearValue()
end

function NameplateCCnTrinket:COMBAT_LOG_EVENT_UNFILTERED()
	local _, combatEvent, _, sourceGUID, _, sourceFlags, _, destGUID, _, destFlags, _, spellID, _, _, AuraType = CombatLogGetCurrentEventInfo()
	--if UnitGUID("target") == destGUID then
	--	print(combatEvent, sourceGUID, destGUID, GetSpellLink(spellID))
	--end
	if isHarm(destFlags) ~= 0 and AuraType == "DEBUFF" then
		if combatEvent ~= "SPELL_AURA_REMOVED" and combatEvent ~= "SPELL_AURA_APPLIED" and combatEvent ~= "SPELL_AURA_REFRESH" then return end

		if not NameplateCCnTrinket.Settings.profile.gSetting.CCShowMonster then
			if strfind(destGUID, "Player") == nil then return end
		end
		if combatEvent == "SPELL_AURA_APPLIED" or combatEvent == "SPELL_AURA_REFRESH" then
			CreateDiminishFrame(destGUID, spellID, true, false)
		elseif combatEvent == "SPELL_AURA_REMOVED" then
			CreateDiminishFrame(destGUID, spellID, false, false)
		end
	end
	if isHarm(sourceFlags) ~= 0 and (combatEvent == "SPELL_CAST_SUCCESS" or combatEvent == "SPELL_AURA_APPLIED" or combatEvent == "SPELL_MISSED" or combatEvent == "SPELL_SUMMON") then
		--[[if sourceGUID == UnitGUID("player") then --and (spellID == 208683 or spellID == 7744) then
			print(combatEvent, sourceGUID, spellID, AuraType)
		end]]
		if CDTimeCache[spellID] then
			if not cooldown[sourceGUID] then cooldown[sourceGUID] = {} end
			if not cooldown[sourceGUID][spellID] then
				CreateBorderTexture(sourceGUID, spellID)
				_G[sourceGUID .. spellID].Texture:SetTexture(CDTextureCache[spellID])
				cooldown[sourceGUID][spellID] = _G[sourceGUID .. spellID]
			end
			_G[sourceGUID .. spellID].c:SetCooldown(GetTime(), CDTimeCache[spellID])
		end
	end
end

function NameplateCCnTrinket:NAME_PLATE_UNIT_ADDED(_, unitID)
	local nameplate = GetNamePlateForUnit(unitID)
	local guid = UnitGUID(unitID)
	npVisible[nameplate] = guid
	npUnitID[guid] = unitID
end

function NameplateCCnTrinket:NAME_PLATE_UNIT_REMOVED(_, unitID)
	local nameplate = GetNamePlateForUnit(unitID)
	local guid = UnitGUID(unitID)
	npVisible[nameplate] = nil
	npUnitID[guid] = nil
end

function NameplateCCnTrinket:OnUpdate(elapsed)
	getUpdate = getUpdate + elapsed

	if getUpdate > UPDATE_INTERVAL then
		UpdateFrame(gradual, CC_MODE)
		UpdateFrame(cooldown, CD_MODE)
		getUpdate = 0
	end
end